# Project options
PROJECT_INIT_CMD_NAME = "init"
PROJECT_INIT_HELP = (
    "[BETA] Used to install Safety Firewall globally, or to initialize a project in the current directory."
    "\nExample: safety init"
)
PROJECT_INIT_DIRECTORY_HELP = (
    "[BETA] Defines a directory for creating a new project. (default: current directory)\n\n"
    "[bold]Example: safety init /path/to/project[/bold]"
)
